@@ -4,13 +4,13 @@ use swc_atoms::js_word;
 use swc_common::{
     pass::{CompilerPass, Repeated},
     util::{move_map::MoveMap, take::Take},
-    Mark, Spanned, DUMMY_SP,
+    Mark, Spanned, SyntaxContext, DUMMY_SP,
 };
 use swc_ecma_ast::*;
 use swc_ecma_transforms_base::pass::RepeatedJsPass;
 use swc_ecma_utils::{
-    extract_var_ids, is_literal, prepend, preserve_effects, undefined, ExprExt, ExprFactory,
-    Hoister, IsEmpty, StmtExt, StmtLike, Value::Known,
+    extract_var_ids, is_literal, prepend_stmt, undefined, ExprCtx, ExprExt, ExprFactory, Hoister,
+    IsEmpty, StmtExt, StmtLike, Value::Known,
 };
 use swc_ecma_visit::{
     as_folder, noop_visit_mut_type, noop_visit_type, Visit, VisitMut, VisitMutWith, VisitWith,
@@ -23,8 +23,15 @@ mod tests;
 /// Not intended for general use. Use [simplifier] instead.
 ///
 /// Ported from `PeepholeRemoveDeadCode` of google closure compiler.
-pub fn dead_branch_remover(_unresolved_mark: Mark) -> impl RepeatedJsPass + VisitMut + 'static {
-    as_folder(Remover::default())
+pub fn dead_branch_remover(unresolved_mark: Mark) -> impl RepeatedJsPass + VisitMut + 'static {
+    as_folder(Remover {
+        changed: false,
+        normal_block: Default::default(),
+        expr_ctx: ExprCtx {
+            unresolved_ctxt: SyntaxContext::empty().apply_mark(unresolved_mark),
+            is_unresolved_ref_safe: false,
+        },
+    })
 }
 
 impl CompilerPass for Remover {
@@ -43,10 +50,12 @@ impl Repeated for Remover {
     }
 }
 
-#[derive(Debug, Default)]
+#[derive(Debug)]
 struct Remover {
     changed: bool,
     normal_block: bool,
+
+    expr_ctx: ExprCtx,
 }
 
 impl VisitMut for Remover {
@@ -140,19 +149,19 @@ impl VisitMut for Remover {
             }
 
             Expr::Cond(cond)
-                if !cond.test.may_have_side_effects()
-                    && (cond.cons.is_undefined()
+                if !cond.test.may_have_side_effects(&self.expr_ctx)
+                    && (cond.cons.is_undefined(&self.expr_ctx)
                         || matches!(*cond.cons, Expr::Unary(UnaryExpr {
                                 op: op!("void"),
                                 ref arg,
                                 ..
-                            }) if !arg.may_have_side_effects()))
-                    && (cond.alt.is_undefined()
+                            }) if !arg.may_have_side_effects(&self.expr_ctx)))
+                    && (cond.alt.is_undefined(&self.expr_ctx)
                         || matches!(*cond.alt, Expr::Unary(UnaryExpr {
                                 op: op!("void"),
                                 ref arg,
                                 ..
-                            }) if !arg.may_have_side_effects())) =>
+                            }) if !arg.may_have_side_effects(&self.expr_ctx))) =>
             {
                 if cfg!(feature = "debug") {
                     debug!("Dropping side-effect-free expressions");
@@ -181,7 +190,7 @@ impl VisitMut for Remover {
         s.visit_mut_children_with(self);
 
         s.init = s.init.take().and_then(|e| match e {
-            VarDeclOrExpr::Expr(e) => ignore_result(*e, true)
+            VarDeclOrExpr::Expr(e) => ignore_result(*e, true, &self.expr_ctx)
                 .map(Box::new)
                 .map(VarDeclOrExpr::from),
             _ => Some(e),
@@ -190,11 +199,11 @@ impl VisitMut for Remover {
         s.update = s
             .update
             .take()
-            .and_then(|e| ignore_result(*e, true).map(Box::new));
+            .and_then(|e| ignore_result(*e, true, &self.expr_ctx).map(Box::new));
 
         s.test = s.test.take().and_then(|e| {
             let span = e.span();
-            if let Known(value) = e.as_pure_bool() {
+            if let Known(value) = e.as_pure_bool(&self.expr_ctx) {
                 return if value {
                     None
                 } else {
@@ -265,7 +274,7 @@ impl VisitMut for Remover {
                 span,
                 key,
                 value: Some(expr),
-            }) if expr.is_undefined()
+            }) if expr.is_undefined(&self.expr_ctx)
                 || match **expr {
                     Expr::Unary(UnaryExpr {
                         op: op!("void"),
@@ -301,7 +310,7 @@ impl VisitMut for Remover {
 
         match p {
             Pat::Assign(assign)
-                if assign.right.is_undefined()
+                if assign.right.is_undefined(&self.expr_ctx)
                     || match *assign.right {
                         Expr::Unary(UnaryExpr {
                             op: op!("void"),
@@ -354,13 +363,13 @@ impl VisitMut for Remover {
                         return Some(e);
                     }
 
-                    ignore_result(*e, true).map(Box::new)
+                    ignore_result(*e, true, &self.expr_ctx).map(Box::new)
                 })
                 .collect()
         } else {
             e.exprs
                 .take()
-                .move_flat_map(|e| ignore_result(*e, false).map(Box::new))
+                .move_flat_map(|e| ignore_result(*e, false, &self.expr_ctx).map(Box::new))
         };
 
         exprs.push(last);
@@ -393,14 +402,16 @@ impl VisitMut for Remover {
                     }
 
                     let mut stmts = vec![];
-                    if let (p, Known(v)) = test.as_bool() {
+                    if let (p, Known(v)) = test.cast_to_bool(&self.expr_ctx) {
                         if cfg!(feature = "debug") {
                             trace!("The condition for if statement is always {}", v);
                         }
 
                         // Preserve effect of the test
                         if !p.is_pure() {
-                            if let Some(expr) = ignore_result(*test, true).map(Box::new) {
+                            if let Some(expr) =
+                                ignore_result(*test, true, &self.expr_ctx).map(Box::new)
+                            {
                                 stmts.push(Stmt::Expr(ExprStmt { span, expr }))
                             }
                         }
@@ -444,7 +455,7 @@ impl VisitMut for Remover {
                         if let Stmt::Empty(..) = *cons {
                             self.changed = true;
 
-                            return if let Some(expr) = ignore_result(*test, true) {
+                            return if let Some(expr) = ignore_result(*test, true, &self.expr_ctx) {
                                 Stmt::Expr(ExprStmt {
                                     span,
                                     expr: Box::new(expr),
@@ -499,7 +510,7 @@ impl VisitMut for Remover {
                     }
 
                     let expr = *expr;
-                    match ignore_result(expr, false) {
+                    match ignore_result(expr, false, &self.expr_ctx) {
                         Some(e) => Stmt::Expr(ExprStmt {
                             span,
                             expr: Box::new(e),
@@ -542,7 +553,7 @@ impl VisitMut for Remover {
 
                         return if let Some(mut finalizer) = finalizer {
                             if let Some(var) = var.map(Decl::from).map(Stmt::from) {
-                                prepend(&mut finalizer.stmts, var);
+                                prepend_stmt(&mut finalizer.stmts, var);
                             }
                             finalizer.into()
                         } else {
@@ -628,10 +639,7 @@ impl VisitMut for Remover {
                         Expr::Lit(Lit::Str(..))
                         | Expr::Lit(Lit::Null(..))
                         | Expr::Lit(Lit::Num(..)) => true,
-                        ref e
-                            if e.is_ident_ref_to(js_word!("NaN"))
-                                || e.is_ident_ref_to(js_word!("undefined")) =>
-                        {
+                        ref e if e.is_nan() || e.is_global_ref_to(&self.expr_ctx, "undefined") => {
                             true
                         }
                         _ => false,
@@ -642,7 +650,7 @@ impl VisitMut for Remover {
                         if cfg!(feature = "debug") {
                             debug!("Removing an empty switch statement");
                         }
-                        return match ignore_result(*s.discriminant, true) {
+                        return match ignore_result(*s.discriminant, true, &self.expr_ctx) {
                             Some(expr) => Stmt::Expr(ExprStmt {
                                 span: s.span,
                                 expr: Box::new(expr),
@@ -661,8 +669,8 @@ impl VisitMut for Remover {
                         }
 
                         let mut stmts = remove_break(s.cases.remove(0).cons);
-                        if let Some(expr) = ignore_result(*s.discriminant, true) {
-                            prepend(&mut stmts, expr.into_stmt());
+                        if let Some(expr) = ignore_result(*s.discriminant, true, &self.expr_ctx) {
+                            prepend_stmt(&mut stmts, expr.into_stmt());
                         }
 
                         let mut block = Stmt::Block(BlockStmt {
@@ -704,8 +712,8 @@ impl VisitMut for Remover {
                                     }
 
                                     _ => {
-                                        if !test.is_ident_ref_to(js_word!("NaN"))
-                                            && !test.is_ident_ref_to(js_word!("undefined"))
+                                        if !test.is_nan()
+                                            && !test.is_global_ref_to(&self.expr_ctx, "undefined")
                                         {
                                             non_constant_case_idx = Some(i);
                                         }
@@ -733,7 +741,9 @@ impl VisitMut for Remover {
                     if let Some(i) = selected {
                         if !has_conditional_stopper(&s.cases[i].cons) {
                             let mut exprs = vec![];
-                            exprs.extend(ignore_result(*s.discriminant, true).map(Box::new));
+                            exprs.extend(
+                                ignore_result(*s.discriminant, true, &self.expr_ctx).map(Box::new),
+                            );
 
                             let mut stmts = s.cases.remove(i).cons;
                             let mut cases = s.cases.drain(i..);
@@ -741,7 +751,7 @@ impl VisitMut for Remover {
                             for case in cases.by_ref() {
                                 exprs.extend(
                                     case.test
-                                        .and_then(|e| ignore_result(*e, true))
+                                        .and_then(|e| ignore_result(*e, true, &self.expr_ctx))
                                         .map(Box::new),
                                 );
 
@@ -760,7 +770,7 @@ impl VisitMut for Remover {
                                 .flat_map(|case| {
                                     exprs.extend(
                                         case.test
-                                            .and_then(|e| ignore_result(*e, true))
+                                            .and_then(|e| ignore_result(*e, true, &self.expr_ctx))
                                             .map(Box::new),
                                     );
                                     case.cons
@@ -775,7 +785,7 @@ impl VisitMut for Remover {
                                 .collect::<Vec<_>>();
 
                             if !decls.is_empty() {
-                                prepend(
+                                prepend_stmt(
                                     &mut stmts,
                                     Stmt::Decl(Decl::Var(VarDecl {
                                         span: DUMMY_SP,
@@ -787,7 +797,7 @@ impl VisitMut for Remover {
                             }
 
                             if !exprs.is_empty() {
-                                prepend(
+                                prepend_stmt(
                                     &mut stmts,
                                     Stmt::Expr(ExprStmt {
                                         span: DUMMY_SP,
@@ -838,7 +848,7 @@ impl VisitMut for Remover {
                                     let mut stmts = remove_break(stmts);
 
                                     if !vars.is_empty() {
-                                        prepend(
+                                        prepend_stmt(
                                             &mut stmts,
                                             Stmt::Decl(Decl::Var(VarDecl {
                                                 span: DUMMY_SP,
@@ -940,20 +950,24 @@ impl VisitMut for Remover {
                             && !has_conditional_stopper(&s.cases.last().unwrap().cons)
                         {
                             let mut exprs = vec![];
-                            exprs.extend(ignore_result(*s.discriminant, true).map(Box::new));
+                            exprs.extend(
+                                ignore_result(*s.discriminant, true, &self.expr_ctx).map(Box::new),
+                            );
 
                             exprs.extend(
                                 s.cases
                                     .iter_mut()
                                     .filter_map(|case| case.test.take())
-                                    .filter_map(|e| ignore_result(*e, true).map(Box::new)),
+                                    .filter_map(|e| {
+                                        ignore_result(*e, true, &self.expr_ctx).map(Box::new)
+                                    }),
                             );
 
                             let stmts = s.cases.pop().unwrap().cons;
                             let mut stmts = remove_break(stmts);
 
                             if !exprs.is_empty() {
-                                prepend(
+                                prepend_stmt(
                                     &mut stmts,
                                     Stmt::Expr(ExprStmt {
                                         span: DUMMY_SP,
@@ -1064,7 +1078,7 @@ impl VisitMut for Remover {
                 }
 
                 Stmt::While(s) => {
-                    if let (purity, Known(v)) = s.test.as_bool() {
+                    if let (purity, Known(v)) = s.test.cast_to_bool(&self.expr_ctx) {
                         if v {
                             if purity.is_pure() {
                                 Stmt::While(WhileStmt {
@@ -1101,7 +1115,7 @@ impl VisitMut for Remover {
                         return Stmt::DoWhile(s);
                     }
 
-                    if let Known(v) = s.test.as_pure_bool() {
+                    if let Known(v) = s.test.as_pure_bool(&self.expr_ctx) {
                         if v {
                             // `for(;;);` is shorter than `do ; while(true);`
                             Stmt::For(ForStmt {
@@ -1115,7 +1129,7 @@ impl VisitMut for Remover {
                             let mut body = prepare_loop_body_for_inlining(*s.body);
                             body.visit_mut_with(self);
 
-                            if let Some(test) = ignore_result(*s.test, true) {
+                            if let Some(test) = ignore_result(*s.test, true, &self.expr_ctx) {
                                 BlockStmt {
                                     span: s.span,
                                     stmts: vec![body, test.into_stmt()],
@@ -1189,7 +1203,7 @@ impl VisitMut for Remover {
 
         if s.cases.iter().all(|case| {
             if let Some(test) = case.test.as_deref() {
-                if test.may_have_side_effects() {
+                if test.may_have_side_effects(&self.expr_ctx) {
                     return false;
                 }
             }
@@ -1321,11 +1335,11 @@ impl Remover {
                             span,
                         }) => {
                             // check if
-                            match test.as_bool() {
+                            match test.cast_to_bool(&self.expr_ctx) {
                                 (purity, Known(val)) => {
                                     self.changed = true;
                                     if !purity.is_pure() {
-                                        let expr = ignore_result(*test, true);
+                                        let expr = ignore_result(*test, true, &self.expr_ctx);
 
                                         if let Some(expr) = expr {
                                             new_stmts.push(T::from_stmt(Stmt::Expr(ExprStmt {
@@ -1384,7 +1398,7 @@ impl Remover {
 ///  - [Some] if `e` has a side effect.
 ///  - [None] if `e` does not have a side effect.
 #[inline(never)]
-fn ignore_result(e: Expr, drop_str_lit: bool) -> Option<Expr> {
+fn ignore_result(e: Expr, drop_str_lit: bool, ctx: &ExprCtx) -> Option<Expr> {
     match e {
         Expr::Lit(Lit::Num(..))
         | Expr::Lit(Lit::Bool(..))
@@ -1393,7 +1407,7 @@ fn ignore_result(e: Expr, drop_str_lit: bool) -> Option<Expr> {
 
         Expr::Lit(Lit::Str(ref v)) if drop_str_lit || v.value.is_empty() => None,
 
-        Expr::Paren(ParenExpr { expr, .. }) => ignore_result(*expr, true),
+        Expr::Paren(ParenExpr { expr, .. }) => ignore_result(*expr, true, ctx),
 
         Expr::Assign(AssignExpr {
             op: op!("="),
@@ -1417,13 +1431,14 @@ fn ignore_result(e: Expr, drop_str_lit: bool) -> Option<Expr> {
             op,
             right,
         }) if op != op!("&&") && op != op!("||") && op != op!("??") => {
-            let left = ignore_result(*left, true);
-            let right = ignore_result(*right, true);
+            let left = ignore_result(*left, true, ctx);
+            let right = ignore_result(*right, true, ctx);
 
             match (left, right) {
                 (Some(l), Some(r)) => ignore_result(
-                    preserve_effects(span, *undefined(span), vec![Box::new(l), Box::new(r)]),
+                    ctx.preserve_effects(span, *undefined(span), vec![Box::new(l), Box::new(r)]),
                     true,
+                    ctx,
                 ),
                 (Some(l), None) => Some(l),
                 (None, Some(r)) => Some(r),
@@ -1438,13 +1453,13 @@ fn ignore_result(e: Expr, drop_str_lit: bool) -> Option<Expr> {
             right,
         }) => {
             if op == op!("&&") {
-                let right = if let Some(right) = ignore_result(*right, true) {
+                let right = if let Some(right) = ignore_result(*right, true, ctx) {
                     Box::new(right)
                 } else {
-                    return ignore_result(*left, true);
+                    return ignore_result(*left, true, ctx);
                 };
 
-                let l = left.as_pure_bool();
+                let l = left.as_pure_bool(ctx);
 
                 if let Known(l) = l {
                     if l {
@@ -1463,16 +1478,16 @@ fn ignore_result(e: Expr, drop_str_lit: bool) -> Option<Expr> {
             } else {
                 debug_assert!(op == op!("||") || op == op!("??"));
 
-                let l = left.as_pure_bool();
+                let l = left.as_pure_bool(ctx);
 
                 if let Known(l) = l {
                     if l {
                         None
                     } else {
-                        ignore_result(*right, true)
+                        ignore_result(*right, true, ctx)
                     }
                 } else {
-                    let right = ignore_result(*right, true);
+                    let right = ignore_result(*right, true, ctx);
                     if let Some(right) = right {
                         Some(Expr::Bin(BinExpr {
                             span,
@@ -1481,7 +1496,7 @@ fn ignore_result(e: Expr, drop_str_lit: bool) -> Option<Expr> {
                             right: Box::new(right),
                         }))
                     } else {
-                        ignore_result(*left, true)
+                        ignore_result(*left, true, ctx)
                     }
                 }
             }
@@ -1506,7 +1521,7 @@ fn ignore_result(e: Expr, drop_str_lit: bool) -> Option<Expr> {
             | op!(unary, "+")
             | op!(unary, "-")
             | op!("!")
-            | op!("~") => ignore_result(*arg, true),
+            | op!("~") => ignore_result(*arg, true, ctx),
             _ => Some(Expr::Unary(UnaryExpr { span, op, arg })),
         },
 
@@ -1521,7 +1536,7 @@ fn ignore_result(e: Expr, drop_str_lit: bool) -> Option<Expr> {
                 }
                 None => None,
                 Some(ExprOrSpread { spread: None, expr }) => {
-                    ignore_result(*expr, true).map(|expr| {
+                    ignore_result(*expr, true, ctx).map(|expr| {
                         Some(ExprOrSpread {
                             spread: None,
                             expr: Box::new(expr),
@@ -1536,12 +1551,13 @@ fn ignore_result(e: Expr, drop_str_lit: bool) -> Option<Expr> {
                 Some(Expr::Array(ArrayLit { span, elems }))
             } else {
                 ignore_result(
-                    preserve_effects(
+                    ctx.preserve_effects(
                         span,
                         *undefined(span),
                         elems.into_iter().map(|v| v.unwrap().expr),
                     ),
                     true,
+                    ctx,
                 )
             }
         }
@@ -1562,12 +1578,13 @@ fn ignore_result(e: Expr, drop_str_lit: bool) -> Option<Expr> {
                 None
             } else {
                 ignore_result(
-                    preserve_effects(
+                    ctx.preserve_effects(
                         span,
                         *undefined(DUMMY_SP),
                         once(Box::new(Expr::Object(ObjectLit { span, props }))),
                     ),
                     true,
+                    ctx,
                 )
             }
         }
@@ -1577,7 +1594,7 @@ fn ignore_result(e: Expr, drop_str_lit: bool) -> Option<Expr> {
             ref callee,
             args,
             ..
-        }) if callee.is_pure_callee() => ignore_result(
+        }) if callee.is_pure_callee(ctx) => ignore_result(
             Expr::Array(ArrayLit {
                 span,
                 elems: args
@@ -1585,6 +1602,7 @@ fn ignore_result(e: Expr, drop_str_lit: bool) -> Option<Expr> {
                     .unwrap_or_else(Default::default),
             }),
             true,
+            ctx,
         ),
 
         Expr::Call(CallExpr {
@@ -1592,26 +1610,31 @@ fn ignore_result(e: Expr, drop_str_lit: bool) -> Option<Expr> {
             callee: Callee::Expr(ref callee),
             args,
             ..
-        }) if callee.is_pure_callee() => ignore_result(
+        }) if callee.is_pure_callee(ctx) => ignore_result(
             Expr::Array(ArrayLit {
                 span,
                 elems: args.into_iter().map(Some).collect(),
             }),
             true,
+            ctx,
         ),
 
-        Expr::Tpl(Tpl { span, exprs, .. }) => {
-            ignore_result(preserve_effects(span, *undefined(span), exprs), true)
-        }
+        Expr::Tpl(Tpl { span, exprs, .. }) => ignore_result(
+            ctx.preserve_effects(span, *undefined(span), exprs),
+            true,
+            ctx,
+        ),
 
         Expr::TaggedTpl(TaggedTpl {
             span,
             tag,
             tpl: Tpl { exprs, .. },
             ..
-        }) if tag.is_pure_callee() => {
-            ignore_result(preserve_effects(span, *undefined(span), exprs), true)
-        }
+        }) if tag.is_pure_callee(ctx) => ignore_result(
+            ctx.preserve_effects(span, *undefined(span), exprs),
+            true,
+            ctx,
+        ),
 
         //
         // Function expressions are useless if they are not used.
@@ -1628,7 +1651,7 @@ fn ignore_result(e: Expr, drop_str_lit: bool) -> Option<Expr> {
                 return None;
             }
 
-            let last = ignore_result(*exprs.pop().unwrap(), true).map(Box::new);
+            let last = ignore_result(*exprs.pop().unwrap(), true, ctx).map(Box::new);
 
             exprs.extend(last);
 
@@ -1645,7 +1668,7 @@ fn ignore_result(e: Expr, drop_str_lit: bool) -> Option<Expr> {
             cons,
             alt,
         }) => {
-            let alt = if let Some(alt) = ignore_result(*alt, true) {
+            let alt = if let Some(alt) = ignore_result(*alt, true, ctx) {
                 alt
             } else {
                 return ignore_result(
@@ -1656,10 +1679,11 @@ fn ignore_result(e: Expr, drop_str_lit: bool) -> Option<Expr> {
                         right: cons,
                     }),
                     true,
+                    ctx,
                 );
             };
 
-            let cons = if let Some(cons) = ignore_result(*cons, true) {
+            let cons = if let Some(cons) = ignore_result(*cons, true, ctx) {
                 cons
             } else {
                 return ignore_result(
@@ -1670,6 +1694,7 @@ fn ignore_result(e: Expr, drop_str_lit: bool) -> Option<Expr> {
                         right: Box::new(alt),
                     }),
                     true,
+                    ctx,
                 );
             };
 
